const { createQuery, convertRequest } = require('./helpers')

test('createQuery: For users, providing an empty constraints object returns the base query', () => {
  const resourceType = 'users'

  // Provide empty constraint object
  const constraints = {}

  // Run function to test
  const query = createQuery(resourceType, constraints)

  // Expect
  expect(query).toEqual({
    path: {
      startsWith: `/${resourceType}/`
    },
    valueField: {
      isAvailable: {
        equals: true
      }
    }
  })
})

test('createQuery: For partners, providing a constraints object w/ appName returns the base query', () => {
  const resourceType = 'partners'
  const appName = 'unitTest'

  // Provide constraint object with appName
  const constraints = {
    appName: {
      equals: appName
    }
  }

  // Run function to test
  const query = createQuery(resourceType, constraints)

  // Expect
  expect(query).toEqual({
    path: {
      startsWith: `/${resourceType}/`
    },
    valueField: {
      isAvailable: {
        equals: true
      },
      'metadata.appName.value': {
        equals: appName
      }
    }
  })
})

test('createQuery: For users, providing a one constraints object returns the correct query', () => {
  const resourceType = 'users'
  const platform = 'link'

  // Provide a constraints object with one constraint
  const constraints = {
    platform: {
      equals: platform
    }
  }

  // Run function to test
  const query = createQuery(resourceType, constraints)

  // Expect
  expect(query).toEqual({
    path: {
      startsWith: `/${resourceType}/`
    },
    valueField: {
      isAvailable: {
        equals: true
      },
      'metadata.platform.value': {
        equals: platform
      }
    }
  })
})

test('createQuery: Providing a two constraints object (with one constraint being "equalsNot") returns the correct query', () => {
  const resourceType = 'partners'
  const appName = 'unitTest'
  const browser = 'chrome'
  // Provide a constraints object with one constraint
  const constraints = {
    appName: {
      equals: appName
    },
    browser: {
      equalsNot: browser
    }
  }

  // Run function to test
  const query = createQuery(resourceType, constraints)

  // Expect
  expect(query).toEqual({
    path: {
      startsWith: `/${resourceType}/`
    },
    valueField: {
      isAvailable: {
        equals: true
      },
      'metadata.appName.value': {
        equals: appName
      },
      'metadata.browser.value': {
        equalsNot: browser
      }
    }
  })
})

test('createQuery: For users, providing a null object returns the base query', () => {
  const resourceType = 'users'

  // Run function to test
  const query = createQuery(resourceType, null)

  // Expect
  expect(query).toEqual({
    path: {
      startsWith: `/${resourceType}/`
    },
    valueField: {
      isAvailable: {
        equals: true
      }
    }
  })
})

test('convertRequest: Providing a request for a user without any constraints returns a converted request with extra data', () => {
  const resourceType = 'users'

  // Provide a request object with one partner requested w/ no constraints
  const rawRequest = {
    requestedResources: {
      users: [{}]
    }
  }

  // Run function to test
  const convertedRequest = convertRequest(rawRequest)

  // Expect
  expect(convertedRequest).toEqual({
    requestedResources: [
      {
        isFulfilled: false,
        query: {
          path: {
            startsWith: `/${resourceType}/`
          },
          valueField: {
            isAvailable: {
              equals: true
            }
          }
        },
        resource: {
          type: resourceType
        }
      }
    ]
  })
})

test('convertRequest: Providing a request without requestedResources throws an error', () => {
  // Provide a request object with no requested resources
  const rawRequest = {}

  // Wrap in a function, otherwise the error will not be caught
  function convertRequestWithNoRequestedResources () {
    // Run function to test
    convertRequest(rawRequest)
  }
  // Expect
  expect(convertRequestWithNoRequestedResources).toThrowError(
    'No requested resources.'
  )
})
