const pMap = require('p-map')
const pReduce = require('p-reduce')

exports.createQueueManager = function () {
  let queue = []

  async function processRequest (db, request) {
    const processedRequestedResources = await pMap(
      request.requestedResources,
      async requestedResource => {
        // Skip if it's already been fulfilled
        if (!requestedResource.isFulfilled) {
          // Get partner based on query
          let matchedResources = await db.get(requestedResource.query)

          // The metadata that is "required" must have been included in the query
          // Filter out the resources that have not been fulfilled
          if (matchedResources && matchedResources.length !== 0) {
            // Get the constraints from the query
            const constraints = Object.keys(
              requestedResource.query.valueField
            ).reduce((acc, constraint) => {
              if (constraint !== 'isAvailable') {
                acc.push(constraint.split('.')[1])
              }
              return acc
            }, [])
            // Remove resources whose required metadata was not all queried
            matchedResources = matchedResources.reduce(
              (acc, matchedResource) => {
                // Get required metadata
                const required = Object.keys(
                  matchedResource.value.metadata
                ).reduce((acc, prop) => {
                  if (matchedResource.value.metadata[prop].required) {
                    acc.push(prop)
                  }
                  return acc
                }, [])
                // Check if all were queried
                const allRequiresFulfilled = required.every(prop =>
                  constraints.includes(prop)
                )
                // Push back into matchedResources if so
                if (allRequiresFulfilled) {
                  acc.push(matchedResource)
                }
                return acc
              },
              []
            )
          }

          // Check if anything was matched
          if (matchedResources && matchedResources.length !== 0) {
            const matchedResource = matchedResources[0]
            // Get id from path (only need the first resource matched)
            const resourceId = matchedResource.path.split(
              `/${requestedResource.resource.type}/`
            )[1]

            // Add the resourceId to the requestedResource
            requestedResource.resource.id = resourceId

            if (requestedResource.resource.type === 'users') {
              // Add the credentials to the requestedResource
              requestedResource.resource.credentials =
                matchedResource.value.credentials
            }

            // Set the availability to false
            await db.set(requestedResource.resource.type, {
              id: resourceId,
              value: {
                ...matchedResource.value,
                isAvailable: false,
                reservedBy: request.from
              }
            })

            // Set isFulfilled to true so it prevents querying it again
            requestedResource.isFulfilled = true
          }
        }
        return requestedResource
      },
      { concurrency: 1 }
    )
    return {
      // Copy the whole request
      ...request,
      // Map over the required resources
      requestedResources: processedRequestedResources
    }
  }

  return {
    push (request) {
      queue.push(request)
    },
    async onResourceProvided (db) {
      console.log('Processing the queue:', queue)
      // Iterate through the queue of requests
      try {
        queue = await pReduce(
          queue,
          async (newQueue, request) => {
            // Process request
            const processedRequest = await processRequest(db, request)

            // Check if all required resources within request are fulfilled
            const requestIsFulfilled = processedRequest.requestedResources.every(
              requestedResource => requestedResource.isFulfilled
            )
            if (requestIsFulfilled) {
              console.log(`Request fulfilled for ${processedRequest.from}`)
              // Once all the partners are reserved, call the callback with the partner IDs
              processedRequest.respond()
            } else {
              // Otherwise, push it back into the queue
              newQueue.push(processedRequest)
            }

            // Return the newQueue for pReduce
            return newQueue
          },
          []
        )
      } catch (e) {
        console.error('Error while processing queue. Queue: ', queue)
        console.error(e)
      }
    },
    async removeFromQueue (peerId) {
      // Iterate through the queue of requests
      queue = await pReduce(
        queue,
        async (newQueue, request) => {
          // Only push back into the queue if it doesn't belong to the peer who disconnected
          if (request.from !== peerId) {
            newQueue.push(request)
          } else {
            // Cancel timeout since the request is being removed
            request.timeout.cancel()
          }
          // Return the newQueue for pReduce
          return newQueue
        },
        []
      )
    }
  }
}
