const { createQueueManager } = require('./queueManager')

function createMockRequest () {
  return {
    requestedResources: [
      {
        isFulfilled: false,
        query: {
          path: {
            startsWith: '/partners/'
          },
          valueField: {
            isAvailable: {
              equals: true
            },
            'metadata.appName.value': {
              equals: 'unitTest'
            }
          }
        },
        resource: {
          type: 'partners'
        }
      }
    ],
    respond: jest.fn(),
    timeout: {
      cancel: jest.fn()
    },
    from: 'mockInstructorId'
  }
}

function createDBGetMockReturnValue () {
  return {
    path: '/partners/mockID',
    value: {
      metadata: {
        appName: 'unitTest',
        required: true
      }
    }
  }
}

// Setup function creates common elements between tests
function setup () {
  return {
    db: {
      get: jest.fn(),
      set: jest.fn()
    },
    request: createMockRequest(),
    dbGetMockReturnValue: createDBGetMockReturnValue()
  }
}

test('onResourceProvided: Providing a request that cannot currently be fulfilled does not invoke respond', async () => {
  // Setup mocks
  const { db, request } = setup()

  // Create queue manager
  const queueManager = createQueueManager()

  // Push dummy request into queue
  queueManager.push(request)

  // Call function that being tested
  await queueManager.onResourceProvided(db)

  // Assertions
  expect(db.set).not.toBeCalled()
  expect(request.respond).not.toBeCalled()
})

test('onResourceProvided: Providing a request that is not fulfilled but then becomes fulfilled from db.get being called invokes respond', async () => {
  // Setup mocks
  const { db, request, dbGetMockReturnValue } = setup()

  // Create queue manager
  const queueManager = createQueueManager()

  // Push dummy request into queue
  queueManager.push(request)

  // Mock return value
  db.get.mockReturnValueOnce([dbGetMockReturnValue])

  // Call function that being tested
  await queueManager.onResourceProvided(db)

  // Assertions
  expect(db.set).toHaveBeenCalledWith('partners', {
    id: 'mockID',
    value: {
      ...dbGetMockReturnValue.value,
      isAvailable: false,
      reservedBy: request.from
    }
  })
  expect(request.respond).toBeCalled()
})

test('onResourceProvided: Providing a request that is not fulfilled then db.get returns an empty array does not invoke respond', async () => {
  // Setup mocks
  const { db, request } = setup()

  // Create queue manager
  const queueManager = createQueueManager()

  // Push dummy request into queue
  queueManager.push(request)

  // Mock return value
  db.get.mockReturnValueOnce([])

  // Call function that being tested
  await queueManager.onResourceProvided(db)

  // Assertions
  expect(db.set).not.toBeCalled()
  expect(request.respond).not.toBeCalled()
})

test('onResourceProvided: Providing two requests that are not fulfilled then a resource becomes available invokes respond for the first request', async () => {
  // Setup mocks
  const { db, dbGetMockReturnValue } = setup()

  // Create queue manager
  const queueManager = createQueueManager()

  // Push dummy request into queue
  const request1 = {
    ...createMockRequest(),
    callbackPath: 'request1'
  }
  queueManager.push(request1)

  // Push a second dummy request into queue
  const request2 = {
    ...createMockRequest(),
    callbackPath: 'request2'
  }
  queueManager.push(request2)

  // Mock return value
  db.get.mockReturnValueOnce([dbGetMockReturnValue])

  // Call function that being tested
  await queueManager.onResourceProvided(db)

  // Assertions
  expect(db.set).toHaveBeenCalledWith('partners', {
    id: 'mockID',
    value: {
      ...dbGetMockReturnValue.value,
      isAvailable: false,
      reservedBy: request1.from
    }
  })
  expect(request1.respond).toBeCalled()
})

test('onResourceProvided: Providing three requests that are not fulfilled then a resource becomes available invokes respond for the first request, and then another resource becomes available invokes respond for the second request', async () => {
  // Setup mocks
  const { db } = setup()

  // Create queue manager
  const queueManager = createQueueManager()

  // Push dummy request into queue
  const request1 = {
    ...createMockRequest(),
    callbackPath: 'request1'
  }
  queueManager.push(request1)

  // Push a second dummy request into queue
  const request2 = {
    ...createMockRequest(),
    callbackPath: 'request2'
  }
  queueManager.push(request2)

  // Push a third dummy request into queue
  const request3 = {
    ...createMockRequest(),
    callbackPath: 'request3'
  }
  queueManager.push(request3)

  // Mock return values

  // First time around when calling onResourceProvided
  const dbGetMockReturnValue1 = {
    ...createDBGetMockReturnValue(),
    path: '/partners/mockID1'
  }

  db.get.mockReturnValueOnce([dbGetMockReturnValue1])
  db.get.mockReturnValueOnce([])
  db.get.mockReturnValueOnce([])

  // Second time around
  const dbGetMockReturnValue2 = {
    ...createDBGetMockReturnValue(),
    path: '/partners/mockID2'
  }

  db.get.mockReturnValueOnce([dbGetMockReturnValue2])

  // Call function that being tested
  await queueManager.onResourceProvided(db)

  // Assertions
  expect(db.set).toHaveBeenLastCalledWith('partners', {
    id: 'mockID1',
    value: {
      ...dbGetMockReturnValue1.value,
      isAvailable: false,
      reservedBy: request1.from
    }
  })
  expect(request1.respond).toBeCalledWith()

  // Call function again that's being tested
  await queueManager.onResourceProvided(db)

  // Assertions
  expect(db.set).toHaveBeenLastCalledWith('partners', {
    id: 'mockID2',
    value: {
      ...dbGetMockReturnValue2.value,
      isAvailable: false,
      reservedBy: request2.from
    }
  })
  expect(request2.respond).toBeCalledWith()
})

test('removeFromQueue: Providing a peerId that disconnected and exists in the queue is removed from the queue', async () => {
  // Setup mocks
  const { db, request } = setup()

  // Create queue manager
  const queueManager = createQueueManager()

  // Push dummy request into queue
  queueManager.push(request)

  // Call function that being tested
  await queueManager.removeFromQueue('mockInstructorId')

  // If the queue is empty (which it should be), then calling onResourceProvided...
  await queueManager.onResourceProvided(db)
  // will not invoke db.get
  expect(db.get).not.toBeCalled()
})

test('removeFromQueue: Providing a peerId that disconnected and does not exist in the queue will not affect the queue', async () => {
  // Setup mocks
  const { db, request } = setup()

  // Create queue manager
  const queueManager = createQueueManager()

  // Push dummy request into queue
  queueManager.push(request)

  // Call function that being tested
  await queueManager.removeFromQueue('notMockInstructorId')

  // If the queue is not empty (since the IDs did not match), then calling onResourceProvided...
  await queueManager.onResourceProvided(db)
  // will invoke db.get
  expect(db.get).toBeCalled()
})
